/**************************************************************************************

 ass.c
 
 Description:
                           
 	Assembler module for Asm45
                           
 History:

	Version 1.0 alpha:	First release

 License information:

	Copyright (c) 2007 Ansgar Kueckes

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
***************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <malloc.h>
#include <math.h>
#include "../include/asm45.h"

/* assembler globals */
int automatic_indirection = 1;
int repeat_opcode = 1;
char module_name[256];
unsigned int module_start = 0;
unsigned int module_size = 0;
int listing_line = 1;

/* assembler forward declarations */
int evaluate_expression(unsigned int address, char *str, int size, Element *ret_value);


/* == assembler routines ============================================================= */

int next_operator(char *str, int size)
{
	int i;
	int level;
	
	if (!str) return -1;
	
	i=size-1;
	level=0;
	do {
		if (str[i] == ')') level++;
		if (str[i] == '(') level--;
		i--;
	} while ((i>=0) && (level > 0));
	if (level>0) {
		return -11;	// error: parenthesis mismatch
	}
	while (i>=0) {
		if ((str[i] == '+') || (str[i] == '-') || (str[i] == '/') || (str[i] == '*')) {
			if (i > 0)
				if ((str[i]=='/') || (str[i]=='*') ||
					((str[i-1] != 'E') && (str[i-1] != 'e'))) return i;	// exponent sign
		}
		i--;
	}
	return 0;
}

int get_integer(char *str, int size, int *ret_int)
{
	int i,end,mult,base,sign,digit,value;
	
	if (!str || !ret_int) return -1;

	value=0;
	mult=1;
	sign=1;
	base=10;
	end = 1;

	if (*str == '$') {
		base = 16;
		str++;
		size--;
	}
	if (*str == '-') {
		sign = -1;
		str++;
		size--;
	}
	
	if (base == 16) {
		while ((end<size) &&
		(((str[end]>='0') && (str[end]<='9')) ||
		 ((str[end]>='A') && (str[end]<='F')) ||
		 ((str[end]>='a') && (str[end]<='f'))))
		 	end++;
		if ((str[end]!=0) && (str[end]!=' ') && (str[end]!= '\t')) return -2;
	}
	else
		while ((end<size) && (str[end]>='0') && (str[end]<='9')) end++;
	if ((end<size) && (base != 16)) {
		if (str[end] == 'B') base = 8;
		else if (str[end] == 'L') base = 2;
		if ((str[end+1]!=0) && (str[end+1]!=' ') && (str[end+1]!= '\t')) return -2;
	}
	for (i=end-1; i>=0; i--) {
		if ((str[i]>='A') && (str[i]<='F')) digit = str[i]-55;
		else if ((str[i]>='a') && (str[i]<='f')) digit = str[i]-87;
		else if ((str[i]>='0') && (str[i]<='9')) digit = str[i]-48;
		else return -12;				// error: illegal digit
		if (base == 2) {
			if (digit>1) return -12;	// error: illegal digit
		} else if (base == 8) {
			if (digit>7) return -12;	// error: illegal digit
		}
		value += digit * mult; mult *= base;
	}
	if (sign == 1) *ret_int = value;
	else *ret_int = 65536 - value;

	return 0;
}

int get_float(char *str, int size, Element *ret_value)
{
	int exponent,digit;
	double value, mult;
	char *end;
	
	if (!str || !ret_value) return -1;

	ret_value->type = ABS;
	ret_value->value = 0;
	ret_value->string = 0;
	ret_value->bcd = 0;
	ret_value->vtype = REAL;
	exponent = 0;
	value = 0;
	mult = 1;

	//printf ("get_float: str=%s size=%d\n", str, size);	

	end = str;
	while (((end-str)<size) && (*end != ' ') && (*end != '\t')) end++;
	end--;
	while (end>=str) {
		if ((*end<'0') || (*end>'9'))
			switch (*end) {
				case '+':	break;
				case '-':	value = -value;
							break;
				case '.':	value = value / mult;
							mult = 1;
							break;
				case 'e':
				case 'E':	exponent = value;
							value = 0;
							mult = 1;
							break;
				case 's':
				case 'S': 	ret_value->vtype = SHORT;
							break;
				default:	return -2;		// syntax error
			}
		else {
			digit = *end-48;
			value = value + (digit * mult);
			mult *= 10;
		}
		end--;
	}
	ret_value->bcd = value * pow(10, exponent);
	//printf ("get_float: ret_value=%+1.11E\n", ret_value->bcd);	
	return 0;
}

void float2bcd(unsigned char *buffer, Element *value)
{
	int exponent;
	char tmp[32];
	
	if (!buffer || !value) return;
		
	switch (value->vtype) {
		// Note:
		// HP 9845 support full precision exponents between -511 and +511, however
		// IEEE754 based machines support a more limited range from -323 up to +308
		case REAL:	sprintf(tmp, "%+1.11E", value->bcd);
					if (tmp[3]=='#') printf("Error: Exponent overflow.\n");
					//printf("float2bcd: tmp=%s\n", tmp);
					memset(buffer, 0, 8);
					exponent = ((tmp[16]-48)*100+(tmp[17]-48)*10+(tmp[18]-48)) & 0x1FF;
					if (tmp[15]=='-') {
						buffer[0]|=0x80;
						exponent=512-exponent;	// two's complement
					}
					if (tmp[0]=='-') buffer[1]|=0x01;
					//printf("float2bcd: exponent=%d\n", exponent);
					buffer[0]|=(exponent/4);
					buffer[1]|=(exponent%4)<<6;
					buffer[2]|=((tmp[1]-48)<<4) | (tmp[3]-48);
					buffer[3]|=((tmp[4]-48)<<4) | (tmp[5]-48);
					buffer[4]|=((tmp[6]-48)<<4) | (tmp[7]-48);
					buffer[5]|=((tmp[8]-48)<<4) | (tmp[9]-48);
					buffer[6]|=((tmp[10]-48)<<4) | (tmp[11]-48);
					buffer[7]|=((tmp[12]-48)<<4) | (tmp[13]-48);
					return;
					
		case SHORT:	sprintf(tmp, "%+1.5E", value->bcd);
					if (tmp[3]=='#') printf("Error: Exponent overflow.\n");
					//printf("float2bcd: tmp=%s\n", tmp);
					memset(buffer, 0, 4);
					exponent = ((tmp[10]-48)*100+(tmp[11]-48)*10+(tmp[12]-48)) & 0x3F;
					if (tmp[9]=='-') {
						buffer[0]|=0x80;
						exponent=64-exponent;	// two's complement
					}
					if (tmp[0]=='-') buffer[0]|=0x01;
					//printf("float2bcd: exponent=%d\n", exponent);
					buffer[0]|=exponent<<1;
					buffer[1]|=((tmp[1]-48)<<4) | (tmp[3]-48);
					buffer[2]|=((tmp[4]-48)<<4) | (tmp[5]-48);
					buffer[3]|=((tmp[6]-48)<<4) | (tmp[7]-48);
					
		default:	return;
	}
}

int lookup_symbol (char* name)
{
	int i;
	
	if (!name) return -1;
	
	for (i=0; i<symbols_size; i++) {
		if (!strcmp(symbol_table[i].name, name))
			return i;
	}
	return -1;	// not found
}

void normalize_symbol(char *name)
{
	char *tmp;
	
	if (!name) return;
		
	tmp = name;
	if (*tmp) {
		if ((*tmp>='a') && (*tmp<='z')) *tmp -= 32;
		tmp++;
		while (*tmp) {
			if ((*tmp>='A') && (*tmp<='Z')) *tmp += 32; 
			tmp++;
		}
	}
	return;
}

void add_symbol(Symbol *symbol)
{
	if (!symbol) return;
	
	if (symbols_size < MAX_SYMBOLS) {
		symbol_table[symbols_size].name = strdup(symbol->name);
		symbol_table[symbols_size].type = symbol->type;
		symbol_table[symbols_size].vtype = symbol->vtype;
		if ((symbol->vtype == STRING) && (symbol_table[symbols_size].string))
			symbol_table[symbols_size].string = strdup(symbol->string);
		symbol_table[symbols_size].value = symbol->value;
		symbols_size++;
	}
	else printf("Error: max. number of symbols exceeded.");
	if (strlen(symbol->name)>longest_symbol_name)
		longest_symbol_name = strlen(symbol->name);
	return;
}

void add_literalspace(unsigned int address, int size)
{
	if (literalspaces_size < MAX_LITERALSPACES) {
		// now fill the free entry
		literalspace_table[literalspaces_size].address = address;
		literalspace_table[literalspaces_size].used = 0;
		literalspace_table[literalspaces_size].list = 0;
		literalspace_table[literalspaces_size].size = size;
		literalspaces_size++;
	}
	else printf("Error: max. number of literal spaces exceeded.");
	return;
}

int lookup_literal(unsigned int address, unsigned char *buffer, int size, unsigned int *ret_address)
{
	int space, i, last;
	Literal *node;
	
	if (!buffer || !ret_address) return -1;
	
	for (space=0; space<literalspaces_size; space++) {
		node = literalspace_table[space].list;
		while (node) {
			// matching size?
			if (node->size == size) {
				last = node->address + size - 1;
				// completely within range?
				if (((node->address < address) && ((address - node->address)<=512)) ||
				   ((node->address >= address) && ((node->address - address)<512))) {
				   	if (((last < address) && ((address - last)<=512)) ||
				   	    ((last >= address) && ((last - address)<512))) {
					   	// matching data?
						for (i=0; i<(size*2); i++)
							if (buffer[i] != node->value[i]) break;
						if (i == (size*2)) {
							*ret_address = node->address;
							return 1;
						}
					}
				}
			}
			node = node->next;
		}
	}
	return 0;
}

int add_literal(unsigned int address, char *buffer, int size, unsigned int *ret_address)
{
	int space, i, position, available, last;
	Literal *tmp,*node;
	
	if (!buffer || !ret_address) return -1;

	// TODO: more intelligent positioning within literal pool
	for (space=0; space<literalspaces_size; space++) {
		position = literalspace_table[space].address + literalspace_table[space].used;
		available = literalspace_table[space].size - literalspace_table[space].used;
		last = position + size - 1;
		// printf("add_literal: address=%d size=%d position=%d availabe=%d last=%d\n",
		//	address, size, position, available, last);
		if (available >= size) 
			if (((position < address) && ((address - position)<=512)) ||
			   ((position >= address) && ((position - address)<512))) {
			   	if (((last < address) && ((address - last)<=512)) ||
			   	    ((last >= address) && ((last - address)<512))) {
					node = malloc(sizeof(Literal));
					node->value = malloc(size*2);
					node->size = size;
					for (i=0; i<(size*2); i++) node->value[i] = buffer[i];
					node->address = position;
					node->next = 0;
		
					tmp = literalspace_table[space].list;
					if (tmp) {
						while (tmp->next) tmp = tmp->next;
						tmp->next = node;
					}
					else literalspace_table[space].list = node;
						
					literalspace_table[space].used += size;
					*ret_address = node->address;
					return 0;
				}
			}
	}
	return -6;	// no usable literal pool found
}

int process_literal(unsigned int address, char *str, Element *ret_value)
{
	Element element;
	char *pos, *end;
	int res, size, literal;
	char expr[256];
	char string[256];
	unsigned char buffer[1024]; // space for literal
	unsigned char *index;
	unsigned int location;
	
	if (!ret_value || !str) return -1;

	/* determine literal declaration type */
	index = buffer;
	literal = 0;
	if (*str == '=') {		// test for literal
		literal = 1;
		str++;
		if (*str == '=') {	// 'complex' literal ?
			literal = 2;
			str++;
		}
	}

	end = str + strlen(str);
	element.string = string;

	do {
		pos = str;
		while (*pos && (*pos != ',')) pos++;
		strncpy(expr, str, pos-str);
		expr[pos-str]=0;

		if ((res = evaluate_expression(address, expr, strlen(expr), &element)) < 0) return res;

		if (element.vtype == STRING) {
			strcpy(index, element.string);
			index += (strlen(element.string));
			if ((index - buffer) % 2) {
				*index = 0;
				index++;
			}
		}
		else if (element.vtype == REAL) {
			float2bcd(index, &element);
			index+=8;
		}
		else if (element.vtype == SHORT) {
			float2bcd(index, &element);
			index+=4;
		}
		else {
			*index = element.value / 256;
			index++;
			*index = element.value % 256;
			index++;
		}
		str = pos + 1;
	} while (str < end);
	size = (index - buffer + 1) / 2;

	if (!lookup_literal(address, buffer, size, &location))	// not found
		if ((res = add_literal(address, buffer, size, &location)) < 0 ) return res;
			
	ret_value->type = REL;
	ret_value->vtype = ADDRESS;
	ret_value->string = 0;
	ret_value->bcd = 0;
	ret_value->value = location;
	return 0;
}

int resolve_symbol (char *str, Element *ret_value)
{
	int i;
	
	if (!str || !ret_value) return -1;
	
	if ((i = lookup_symbol(str)) < 0) return -10;	// error: undefined symbol
	ret_value->type = symbol_table[i].type;
	ret_value->vtype = symbol_table[i].vtype;
	ret_value->value = symbol_table[i].value;
	ret_value->bcd = symbol_table[i].bcd;
	ret_value->string = symbol_table[i].string;
	return 0;
}

int resolve_element(unsigned int address, char *str, int size, Element *ret_value)
{
	int start,end,res,i;
	int value;
	char tmp[256];
	
	if (!str || !ret_value) return -1;
	
	value=0;
	start=0;
	
	/* subexpression */
	if (str[start]=='(') {
		/* strip surrounding pairs of parenthesis */
		if (str[start+size-1] == ')')
			{ start++; size -= 2; }
		else return -11;				// parenthesis error
		return evaluate_expression (address, str+start, size, ret_value);
	}

	/* literal */
	if (str[start]=='=') {
		strncpy(tmp, str+start, size);
		tmp[size]=0;
		if ((res = process_literal(address, tmp, ret_value)) < 0) return res;
		return 0;
	}

	/* string */
	// TODO: escape control (control characters & double quotes), e.g. \020 \" \n \t etc.
	if (str[start]=='\"') {
		start++;
		end = start+1;
		while ((end < size) && (str[end] != '\"')) end++;
		if (ret_value->string) strncpy(ret_value->string, str+start, end-start);
		ret_value->string[end-start]=0;
		ret_value->type = ABS;
		ret_value->vtype = STRING;
		return 0;
	}

	/* ascii char */
	if (str[start]=='\'') {
		ret_value->type = ABS;
		ret_value->vtype = ASCII;
		ret_value->value = str[start+1];
		return 0;
	}
	
	/* current address identifier */
	if (str[start]=='*') {
		ret_value->type = REL;
		ret_value->vtype = ADDRESS;
		ret_value->value = address;
		return 0;
	}
	
	/* symbol */
	if (((str[start]>='A') && (str[start]<='Z')) ||
		((str[start]>='a') && (str[start]<='z'))) {
		end = start+1;
		while ((end<size) &&
		(((str[end]>='A') && (str[end]<='Z')) ||
		 ((str[end]>='a') && (str[end]<='z')) ||
		 ((str[end]>='0') && (str[end]<='9')) ||
		  (str[end]=='_'))) end++;
		if ((end-start)>MAX_LEN_SYMBOL) return -13; // error: symbol name maxlen exceeded
		else {
			//normalize_symbol(str+start, end-start);
			strncpy(tmp, str+start, end-start);
			tmp[end-start]=0;
			
			if ((i = lookup_symbol(tmp)) < 0) return -10;	// error: undefined symbol
				
			ret_value->type = symbol_table[i].type;
			ret_value->vtype = symbol_table[i].vtype;
			ret_value->value = symbol_table[i].value;
			ret_value->bcd = symbol_table[i].bcd;
			ret_value->string = symbol_table[i].string;
			
			return 0;
			// return resolve_symbol (tmp, ret_value);
		}
	}

	/* integer & floating point */
	if ((str[start]=='$') ||(str[start]=='+') || (str[start]=='-') || ((str[start]>='0') && (str[start]<='9'))) {
		end = start;
		while ((end<size) && (str[end]!='.')) end++;
		if (str[end]=='.') {
			return get_float(str+start, size, ret_value);
		} else {		
			res = get_integer(str+start, size, &value);
			ret_value->type = ABS;
			ret_value->vtype = INTEGER;
			ret_value->value = value;
		}
		return res;
	}
	
	/* error */
	return -5;
}

int evaluate_expression(unsigned int address, char *str, int size, Element *ret_value)
{
	int pos, result, res1, res2;
	Element op1, op2, tmp;
	char op1_string[256];
	char op2_string[256];
	
	if (!str || (size < 0) || !ret_value) return -1;
	
	op1.string = op1_string;
	op2.string = op2_string;
	pos=0;
	result=0;
	
	/* process operator */
    pos = next_operator(str, size);
    if (pos < 0) return pos;	// error
    if (pos == 0) {				// no operator
		return resolve_element(address, str, size, ret_value);
    }

	res2 = evaluate_expression (address, str+pos+1, size-pos-1, &op2);
	res1 = evaluate_expression (address, str, pos, &op1);
	if (res1<0) return res1;	// error
	if (res2<0) return res2;	// error
		
	/* string concatenation */
	if ((op1.vtype == STRING) && (op2.vtype == STRING) && (str[pos] == '+')) {
		ret_value->type = ABS;
		ret_value->vtype = STRING;
		ret_value->value = 0;
		ret_value->string = malloc(strlen(op1.string)+strlen(op2.string)+1);
		strcpy(ret_value->string, op1.string);
		strcat(ret_value->string, op2.string);
		return result;		
	}
	if ((op1.vtype == STRING) || (op2.vtype == STRING)) return -14;	// error
		
	/* floating point arithmetics */
	if (((op1.vtype == REAL) || (op1.vtype == SHORT)) && (op2.type == ABS)) {
		switch (str[pos]) {
			case '+': if ((op2.vtype == REAL) || (op2.vtype == SHORT)) tmp.bcd = op1.bcd + op2.bcd;
					  else tmp.bcd = op1.bcd + op2.value;
					  break;
			case '-': if ((op2.vtype == REAL) || (op2.vtype == SHORT)) tmp.bcd = op1.bcd - op2.bcd;
					  else tmp.bcd = op1.bcd - op2.value;
					  break;
			case '/': if ((op2.vtype == REAL) || (op2.vtype == SHORT)) {
							if (op2.bcd == 0) return -15;			// error: divide by zero
							else tmp.bcd = op1.bcd / op2.bcd;
					  }
					  else {
						if (op2.value == 0) return -15;			// error: divide by zero
					  	tmp.bcd = op1.bcd / op2.value;
					  }
					  break;
			case '*': if ((op2.vtype == REAL) || (op2.vtype == SHORT)) tmp.bcd = op1.bcd * op2.bcd;
					  else tmp.bcd = op1.bcd * op2.value;
					  break;
			default: return -17;	// error: illegal operator
		}
		ret_value->type = ABS;
		if ((op1.vtype == REAL) || (op2.vtype == REAL))
			ret_value->vtype = REAL;
		else
			ret_value->vtype = SHORT;
		ret_value->value = 0;
		ret_value->string = 0;
		ret_value->bcd = tmp.bcd;
		return result;		
	}
	else if (((op2.vtype == REAL) || (op2.vtype == SHORT)) && (op1.type == ABS)) {
		switch (str[pos]) {
			case '+': tmp.bcd = op1.value + op2.bcd;
					  break;
			case '-': tmp.bcd = op1.value - op2.bcd;
					  break;
			case '/': if (op2.bcd == 0) return -15;			// error: divide by zero
					  tmp.bcd = op1.value / op2.bcd;
					  break;
			case '*': tmp.bcd = op1.value * op2.bcd;
					  break;
			default: return -17;	// error: illegal operator
		}
		ret_value->type = op2.type;
		ret_value->vtype = op2.vtype;
		ret_value->value = 0;
		ret_value->string = 0;
		ret_value->bcd = tmp.bcd;
		return result;		
	}
	if ((op1.vtype == REAL) || (op1.vtype == SHORT) ||
		(op2.vtype == REAL) || (op2.vtype == SHORT)) return -16;
		
	/* integer arithmetics and combination rules */
	switch (str[pos]) {
		case '+': if ((op1.type == ABS) && (op2.type == ABS)) tmp.type = ABS;
			  else if ((op1.type == ABS) && (op2.type == REL)) tmp.type = REL;
			  else if ((op1.type == REL) && (op2.type == ABS)) tmp.type = REL;
			  else return -16;	// error: illegal combination
			  tmp.value = op1.value + op2.value;
			  break;
		case '-': if ((op1.type == ABS) && (op2.type == ABS)) tmp.type = ABS;
			  else if ((op1.type == REL) && (op2.type == ABS)) tmp.type = REL;
			  else if ((op1.type == REL) && (op2.type == REL)) tmp.type = ABS;
			  else return -16;	// error: illegal combination
			  tmp.value = op1.value - op2.value;
			  break;
		case '/': if ((op1.type == ABS) && (op2.type == ABS)) tmp.type = ABS;
			  else return -16;	// error: illegal combination
			  if (op2.value == 0) return -15; // error: divide by zero
			  else tmp.value = op1.value / op2.value;
			  break;
		case '*': if ((op1.type == ABS) && (op2.type == ABS)) tmp.type = ABS;
			  else return -16;	// error: illegal combination
			  tmp.value = op1.value * op2.value;
			  break;
		default: return -17;	// error: illegal operator
	}
	tmp.vtype = INTEGER;
	/*
	printf("%s = %d,", str, tmp.value);
	switch (tmp.type) {
		case ABS: printf("ABS\n"); break;
		case REL: printf("REL\n"); break;
		case EXT: printf("EXT\n"); break;
		default: printf("type unknown\n");
	}
	*/
	ret_value->type = tmp.type;
	ret_value->vtype = tmp.vtype;
	ret_value->value = tmp.value;
	ret_value->string = 0;
	return result;
}

void parse_line(char *str, char **ret_symbol, char **ret_opcode, char **ret_operand, char **ret_comment)
{
	char *tmp1, *tmp2, *end;
	int i, symbol, quote, hexint;
	
	if (!str || !ret_symbol || !ret_opcode || !ret_operand || !ret_comment) return;

	*ret_symbol = 0;
	*ret_opcode = 0;
	*ret_operand = 0;
	*ret_comment = 0;
	symbol = 0;
	quote = 0;
	hexint = 0;
	
	end = str + strlen(str);
	tmp1 = tmp2 = str;
	
	/* ASMA full line comment? */
	if (*tmp1 == '*') {
		*ret_comment = tmp1+1;
		return;
	}
	
	while (tmp2<=end) {
		if (*ret_opcode && !quote) {
			if (((*tmp2>='A') && (*tmp2<='Z')) || ((*tmp2>='a') && (*tmp2<='z'))) {
				if (!symbol && !hexint) {
					// first letter of symbol or solitaire
					if ((*tmp2>='a') && (*tmp2<='z')) *tmp2 -= 32;
					symbol = 1;
				} else {
					if (hexint && (((*tmp2>='G') && (*tmp2<='Z')) || ((*tmp2>='g') && (*tmp2<='z'))))
						hexint = 0;
					else
						if ((*tmp2>='A') && (*tmp2<='Z')) *tmp2 += 32;
				}
			} else if ((*tmp2<'0') || (*tmp2>'9')) {
				if (*tmp2 != '_') symbol = 0;
				hexint = 0;
			}
		}
		quote = 0;
		switch (*tmp2) {
			case '!':
			case ';':	*ret_comment = tmp2+1;	
			case 0:		if (*ret_opcode)
							*ret_operand = tmp1;				// operand
						else {
							if ((tmp2 - tmp1) == 3) {
								for (i=0; i<3; i++)
									if ((tmp1[i]<'A') || (tmp1[i]>'Z')) break;
								if ((i == 3) || (!strncmp(tmp1, "???", 3))) {
									*ret_opcode = tmp1;			// opcode
								}
							}
						}
						*tmp2 = 0;
						return;									// done
			case ':':	if (!*ret_opcode) {
							*ret_symbol = tmp1;					// symbol
							*tmp2++ = 0;
							tmp1 = tmp2;
						}
						break;
			case ' ':
			case '\t':	if (!*ret_opcode) {
							if ((tmp2 - tmp1) == 3) {
								for (i=0; i<3; i++)
									if ((tmp1[i]<'A') || (tmp1[i]>'Z')) break;
								if ((i == 3) || (!strncmp(tmp1, "???", 3))) {
									*ret_opcode = tmp1;			// opcode
									*tmp2 = 0;
								}
							}
							tmp2++;
							tmp1 = tmp2;
						} else tmp2++;
						break;
			case '\"':	if (*ret_opcode) {
							tmp2++;
							while (*tmp2 && (*tmp2 != '\"')) tmp2++;
						}
						if (*tmp2 == '\"') tmp2++;
						break;
			case '$':	hexint = 1;
						tmp2++;
						break;
			case '\'':	// for ascii output from disassembler
						if (!*ret_symbol && !*ret_opcode) {
							if (*tmp2) tmp2++;
							if (*tmp2) tmp2++;
							if (*tmp2) tmp2++;
						}
						else quote = 1;
			default:	if (*tmp2) tmp2++;
		}
	}
	return;
}


/* data generator routines */

int write_word(FILE* ofp, unsigned int value)
{
	unsigned char buffer[2];
	int res;
	
	if (!ofp) return 0;

	buffer[0]=value / 256;
	buffer[1]=value % 256;
	res = fwrite (buffer, 2, 1, ofp);
	if (res<1) {
		fprintf(stderr, "Error: could not write to file.\n");
		return -1;
	}
	return 0;
}

void line_out (char *out, char *comment)
{
	int i;
	
	printf("%s", out);
	for (i=35+longest_symbol_name; i>strlen(out); i--) printf(" ");
	if (comment) printf(" ;%s", comment);
	printf ("\n");
	return;
}

void word_out(FILE* ofp, unsigned int address, unsigned int value, char* label, char *opcode, char *operand, char *comment)
{
	char line[1024];
	
	value &= 0xffff;
	write_word(ofp, value);
	
	*line = 0;
	if (!listing) return;
	if (numbers) printf_string(line, "%3d ", listing_line++);
	if (octal)
		printf_string(line, "%06o %06o",address, value);
	else
		printf_string(line, "%04x %04x",address, value);
	if (label) printf_string(line, " %s:%s", label,
		"               "+15-longest_symbol_name+strlen(label));
	else printf_string(line, "                 "+15-longest_symbol_name);
	if (opcode) printf_string(line, " %s", opcode);
	if (operand) printf_string(line, " %s", operand);
	line_out(line, comment);
	return;
}

void string_out(FILE *ofp, unsigned int address, char *str, char *label, char *opcode, char *operand, char *comment)
{
	int i,len;
	unsigned int out, count;
	unsigned char *tmp;
	char line[1024];
	
	if (!str) return;
		
	count = address;
	len=strlen(str);
	tmp = str;
	for (i=0; i<len; i+=2) {
		out = tmp[i]*256+tmp[i+1];
		write_word(ofp, out);
		if (listing) {
			*line = 0;
			if (numbers) printf_string(line, "%3d ", listing_line++);
			if (octal)
				printf_string(line, "%06o %06o",count++, out);
			else
				printf_string(line, "%04x %04x",count++, out);				
			if (i==0) {
				if (label) printf_string(line, " %s:%s", label,
					"               "+15-longest_symbol_name+strlen(label));
				else printf_string(line, "                 "+15-longest_symbol_name);
				if (opcode) printf_string(line, " %s", opcode);
				if (operand) printf_string(line, " %s", operand);
				line_out(line, comment);
			} else
				line_out(line, 0);
		}
	}
	return;
}

void bcd_out(FILE *ofp, unsigned int address, Element *value, char *label, char *opcode, char *operand, char *comment)
{
	int i;
	unsigned int out, count;
	unsigned char tmp[8];
	char line[1024];
	
	count = address;
	float2bcd(tmp, value);
	if (value->vtype == REAL) {
		for (i=0;i<8; i+=2) {
			out = tmp[i]*256+tmp[i+1];
			write_word(ofp, out);
			if (listing) {
				*line = 0;
				if (numbers) printf_string(line, "%3d ", listing_line++);
				if (octal)
					printf_string(line, "%06o %06o",count++, out);
				else
					printf_string(line, "%04x %04x",count++, out);					
				if (i==0) {
					if (label) printf_string(line, " %s:%s", label,
						"               "+15-longest_symbol_name+strlen(label));
					else printf_string(line, "                 "+15-longest_symbol_name);
					if (opcode) printf_string(line, " %s", opcode);
					if (operand) printf_string(line, " %s", operand);
					line_out(line, comment);
				} else
					line_out(line, 0);

			}
		}
	}
	else if (value->vtype == SHORT) {
		for (i=0;i<4; i+=2) {
			out = tmp[i]*256+tmp[i+1];
			write_word(ofp, out);
			if (listing) {
				*line = 0;
				if (numbers) printf_string(line, "%3d ", listing_line++);
				if (octal)
					printf_string(line, "%06o %06o",count++, out);
				else
					printf_string(line, "%04x %04x",count++, out);					
				if (i==0) {
					if (label) printf_string(line, " %s:%s", label,
						"               "+15-longest_symbol_name+strlen(label));
					else printf_string(line, "                 "+15-longest_symbol_name);
					if (opcode) printf_string(line, " %s", opcode);
					if (operand) printf_string(line, " %s", operand);
					line_out(line, comment);
				} else
					line_out(line, 0);
			}
		}
	}
	return;
}

void literalspace_out(FILE* ofp, unsigned int address, int complete, char *label, char *opcode, char *operand, char *comment)
{
	int space, i, count;
	unsigned int out;
	Literal *node;
	char line[1024];
	
	count = address;
	for (space=0; space<literalspaces_size; space++)
		if (literalspace_table[space].address == address) {
			node = literalspace_table[space].list;
			while (node) {
				for (i=0; i<node->size; i++) {
					out = node->value[2*i]*256+node->value[2*i+1];
					write_word(ofp, out);
					if (listing) {
						*line = 0;
						if (numbers) printf_string(line, "%3d ", listing_line++);
						if (octal)
							printf_string(line, "%06o %06o", count, out);
						else
							printf_string(line, "%04x %04x", count, out);						
						if (count == address) {
							if (label) printf_string(line, " %s:%s", label,
								"               "+15-longest_symbol_name+strlen(label));
							else printf_string(line, "                 "+15-longest_symbol_name);
							if (opcode) printf_string(line, " %s", opcode);
							if (operand) printf_string(line, " %s", operand);
//							line_out(line, "; LIT ");
							line_out(line, comment);
						} else
							line_out(line, 0);
					}
					count++;
				}
				node = node->next;
			}
			if (complete)
				while ((count-address) < literalspace_table[space].size) {
					write_word(ofp, 0);
					if (listing) {
						*line = 0;
						if (numbers) printf_string(line, "%3d ", listing_line++);
						if (octal)
							printf_string(line, "%06o %06o", count, 0);
						else
							printf_string(line, "%04x %04x", count, 0);						
						if (count == address) {
							if (label) printf_string(line, " %s:%s", label,
								"               "+15-longest_symbol_name+strlen(label));
							else printf_string(line, "                 "+15-longest_symbol_name);
							if (opcode) printf_string(line, " %s", opcode);
							if (operand) printf_string(line, " %s", operand);
							line_out(line, comment);
						} else
							line_out(line, 0);
					}
					count++;
				}
		}
	return;
}

int data_generator(FILE *ofp, int pass, unsigned int *address, char *label, char *opcode, char *operand, char *comment)
{
	Element element;
	char *start, *pos, *end;
	int res;
	char expr[256];
	char string[256];
	
	if (!address || !operand) return -1;

	start = operand;
	end = start + strlen(start);
	element.string = string;
	
	/* literal */
	if (*start=='=') {
		if (pass>1)
			if ((res = process_literal(*address, start, &element)) < 0) return res;
		if (pass == 3) {
			word_out(ofp, *address, element.value, label, opcode, operand, comment);
			label = opcode = operand = 0;
		}
		(*address)++;
	} else
		
	/* all others */
	do {
		pos = start;
		while (*pos && (*pos != ',')) pos++;
		strncpy(expr, start, pos-start);
		expr[pos-start]=0;
		if ((res = evaluate_expression(*address, expr, strlen(expr), &element)) < 0) {
			if ((pass>1) || (res!=-10)) return res;
		}
		if (element.vtype == STRING) {
			if (pass == 3) {
				string_out(ofp, *address, element.string, label, opcode, operand, comment);
				label = opcode = operand = 0;
			}
			*address += ((strlen(element.string)+1) / 2);
		}
		else if (element.vtype == REAL) {
			if (pass == 3) {
				bcd_out(ofp, *address, &element, label, opcode, operand, comment);
				label = opcode = operand = 0;
			}	
			*address += 4;
		}
		else if (element.vtype == SHORT) {
			if (pass == 3) {
				bcd_out(ofp, *address, &element, label, opcode, operand, comment);
				label = opcode = operand = 0;
			}
			*address += 2;
		}
		else {
			if (pass == 3) {
				word_out(ofp, *address, element.value, label, opcode, operand, comment);
				label = opcode = operand = 0;
			}
			(*address)++;
		}
		start = pos + 1;
		comment = 0;
	} while (start < end);
	return 0;
}

/* code generator routines */

int code_generator(FILE *ofp, int pass, unsigned int address, char *label, char *opcode, char *operand, char *comment, int line)
{
	int i,size,res;
	unsigned int machine_code = 0;
	unsigned int location;
	unsigned char buffer[2];
	char tmp[256];
	char modifier;
	Op_code *current = op_table;
	Element element;
	
	if (!opcode) return -1;

	// lookup opcode	
	for (i=0; i < op_table_size; i++, current++)
	   	if (!strcmp(opcode, current->opcode)) break;
	if (i == op_table_size) return -18;								// error: improper opcode
		
	if (pass == 1) return 0;

	if (!operand) {
		if (current->type > 0) return -22;							// error: missing operand
	}
	else if (operand) {
		if (current->type > 0) {
			if (strlen(operand) == 0) return -22;

			strcpy(tmp, operand);
			size = strlen(tmp);
		
			// extract modifier for indirect, set, clear, increment, decrement
			modifier=0;
			if (size>1)
				if (tmp[0] != '=')
					if ((tmp[size-2]==',' &&
					   ((tmp[size-1]=='I') || (tmp[size-1]=='C') ||
					    (tmp[size-1]=='S') || (tmp[size-1]=='D')))) {
						tmp[size-2]=0;
						modifier=tmp[size-1];
						size-=2;
					}
		
			if (operand[0] == '=') {
				if ((res = process_literal(address, tmp, &element)) < 0) return res;
			} else {
				if ((res = evaluate_expression(address, tmp, size, &element)) < 0) return res;
			}
		}
	}

	//printf("### current->type=%d modifier=<%c>\n", current->type, modifier);
	
	machine_code = current->pattern;								// base opcode pattern
		
	// defaults & modifiers bits
	switch (current->type) {
		case 1:	// location
		case 2:	// indirect location
					// Note: not really relevant, decision is done later with respect to offset
					if (element.type == REL)
						machine_code |= 0x0400;						// current page
					else if (element.type == ABS)					
						machine_code &= 0xFBFF;						// base page
					else return -19;								// TODO: external
		case 3: // register
		case 4: // indirect register
				if (modifier == 'I') machine_code |= 0x8000;		// indirect
				else machine_code &= 0x7FFF;						// direct
				break;
		case 8:	// skip
		case 9:	// skip with set
		case 10:// skip with clear
		case 15:// skip with hold
				if (modifier == 'C') {
					if (current->type == 8) return -28;				// no modifier allowed
					machine_code |= 0x0080;							// no hold
					machine_code &= 0xFFBF;							// clear
				}
				else if (modifier == 'S') {
					if (current->type == 8) return -28;				// no modifier allowed
					machine_code |= 0x00C0;							// no hold & set
				}
				else if (current->type != 8)
					machine_code &= 0xFF5F;							// clear hold bit & clear
				break;
		case 12:// stack
		case 13:// stack with increment
		case 14:// stack with decrement
				if (modifier == 'D') machine_code |= 0x0080;		// decrement
				else if (modifier == 'I') machine_code &= 0xFF7F;	// increment
				else if (opcode[0]=='W') machine_code |= 0x0080;	// decrement
				else machine_code &= 0xFF7F;						// increment
				break;
		default: ;
	}
	
	// check & copy in operands
	offset = element.value - address;
	switch (current->type) {
		case 1:	// location
		case 2:	// indirect location
				// base page addressing
				if (((element.value>=65024) || (element.value<512)) && (element.type == ABS)) {
					machine_code &= 0xFBFF;		// clear bit
					if (element.value<512)
						machine_code |= element.value;
					else 
						machine_code |= element.value & 0x3FF;
					//if ((element.type == REL) && (pass == 3))
					//	fprintf(stderr, "Line %d: Warning: base page address evaluates as relocatable.\n", line);
					break;
				}
				// current page addressing
				else if ((offset>=-512) && (offset<512)) {
					if (offset<0 ) offset = 1024 + offset;
					machine_code |= 0x0400;		// set bit
					machine_code |= offset;
					if ((element.type == ABS) && (pass == 3))
						fprintf(stderr, "Line %d: Warning: current page address evaluates as absolute.\n", line);
					break;
				}
				// try automatic indirection to avoid out-of-range situation
				else if (!(machine_code & 0x8000) && automatic_indirection) {
					// create vector
					buffer[0] = element.value / 256;
					buffer[1] = element.value % 256;
					if (!lookup_literal(address, buffer, 1, &location))
						if ((res = add_literal(address, buffer, 1, &location)) < 0 ) return res;
					offset = location - address;
					if (offset<0 ) offset = 1024 + offset;
					machine_code |= offset | 0x8000;		// paste in vector & change to indirect
					// sprintf(tmp, "*%+d,I  ; -automatic indirection-", location - address);
					if (octal)
						sprintf(tmp, "%06oB,I  ; -automatic indirection (%06oB)-", location, element.value);
					else
						sprintf(tmp, "$%04x,I  ; -automatic indirection ($%04x)-", location,  element.value);						
					operand = tmp;
				}
				else return -7;	// out of range
				break;
		case 3: // register
		case 4: // indirect register
				if (element.type == ABS) {
					if ((element.value>=0) && (element.value<16))
						machine_code |= element.value;
					else return -7;	// out of range
				} else return -9;	// type not allowed
				break;
		case 5: // value
		case 6: // shift
		case 7: // complemented shift
				if (element.type == ABS) {
					if ((element.value>0) && (element.value<17)) {
						if (current->type == 7)
							machine_code |= 15 - element.value;		// TODO: check in specs
						else
							machine_code |= element.value - 1;
					}
					else return -7;	// out of range
				} else return -9;	// type not allowed
				break;
		case 8:	// skip
		case 9:	// skip with set
		case 10:// skip with clear
		case 15:// skip with hold
		case 11:// complemented skip (RET instruction)
				if (current->type == 11) {
					if (element.type == ABS) {
						i = element.value;
						if ((i>=-32) && (i<32)) {
							if (i<0) i = 64 + i;
							machine_code |= i;
						} else return -7;	// out of range
					}
					else return -9;		// type not allowed
				}
				else {
					if (element.type == REL) {
						if ((offset>=-32) && (offset<32)) {
							if (offset<0 ) offset = 64 + offset;
							machine_code |= offset;
						} else return -7;	// out of range
					}
					else return -9;		// type not allowed
				}
				break;
		case 12:// stack
		case 13:// stack with increment
		case 14:// stack with decrement
				if (element.type == ABS) {
					if ((element.value>=0) && (element.value<8))
						machine_code |= element.value;
					else return -7;	// out of range
				} else return -9;	// type not allowed
				break;
		
		default: ;

	}

	if (pass == 3) word_out(ofp, address, machine_code, label, opcode, operand, comment);

	return 0;
}


/* 	multipass assembler routine:

	pass 1:	collect symbol definitions and allocate space for data & literals
	pass 2: collect & resolve all expressions & literals
	pass 3: generate machine code for data, instructions & literals
*/

int assemble_line(FILE *ofp, int pass, unsigned int *address, char *str, int line)
{
	char *label, *opcode, *operand, *comment, *tmp;
	char string[256];
	int i, res, size, tmp_repeat_opcode, tmp_label_size;
	unsigned int location;
	Element element;
	Symbol symbol;
	
	if (!address || !str) return -1;

	element.string = string;
	tmp_repeat_opcode = 0;

	parse_line (str, &label, &opcode, &operand, &comment);

	if (label)	{
		// normalize
		tmp_label_size = strlen(label);
		normalize_symbol(label);
		if (tmp_label_size>MAX_LEN_SYMBOL) {
			label[MAX_LEN_SYMBOL] = 0;
			if (pass == 3) fprintf(stderr, "Line %d: Warning: max. symbol length exceeded, symbol name truncated.\n", line);
		}
		// if (pass == 1) printf("label=%s ", label);
		if ((res = lookup_symbol(label)) >= 0) {
			if (opcode)
				if (strcmp(opcode, "EQU") && strcmp(opcode, "SET"))
					if (symbol_table[res].value != *address) {
						//printf("label=%s label_address=%04x current_address=%04x\n", label, symbol_table[res].value, *address);
						return -3;
					}
		}
		else {
			symbol.name = label;
			symbol.value = *address;
			symbol.vtype = ADDRESS;
			if (base_page_mode && ((*address < 512) || (*address >= 65024)))
				symbol.type = ABS;
			else
				symbol.type = REL;
			symbol.string = 0;
			add_symbol(&symbol);
		}
	}
	while (opcode && repeat_opcode) {
		// if (pass==1) printf("address=%06o opcode=%s", *address, opcode);
		if (operand) {
			// normalize
			tmp = operand;
			i=0;
			while (*tmp) {
				if (*tmp == '\"') {
					operand[i++] = *tmp++;
					while (*tmp && (*tmp != '\"'))
						operand[i++] = *tmp++;
					if (*tmp == 0) return -20;	// quoting error
					else operand[i++] = *tmp++;
				}
				else if (*tmp == '[') {
					tmp++;
					while (*tmp && (*tmp != ']'))
						tmp++;
					if (*tmp == 0) return -29;	// bracket error
					else tmp++;
				}
				else if ((*tmp != ' ') && (*tmp != '\t')) {
					operand[i++] = *tmp++;
				}
				else tmp++;
			}
			operand[i]=0;
			// if (pass==1) printf(" operand=%s", operand);
		}
		
		// if (pass==1) printf("\n");
		
		// pseudo opcodes
		if (!strcmp(opcode, "???")) {
			// to be compatible to invalid opcode (probably data) in disassembly.
			// similar to DAT, but only one single word of input/output.
			if (pass == 3) {
				if ((res = get_integer (operand, strlen(operand), &location)) < 0)
					word_out(ofp, *address, 0, label, opcode, 0, comment);				// relax
				else
					word_out(ofp, *address, location, label, opcode, operand, comment);	// word out
			}
			(*address)++;
		}
		else if (!strcmp(opcode, "NAM")) {	// module name
			if (repeat_opcode > 1) return -21;
			if (!operand) return -22;
			if ((pass == 1) && *module_name) return -23;		// already declared
			tmp = operand;
			while (*tmp && (*tmp!=' ') && (*tmp!='\t')) tmp++;	
			*tmp=0;
			if ((tmp==operand) || ((tmp-operand)>MAX_LEN_SYMBOL)) return -24;
			normalize_symbol(operand);
			strcpy(module_name, operand);
			module_start = *address;
		}
		else if (!strcmp(opcode, "END")) {	// end of module
			if (repeat_opcode > 1) return -21;
			if (!operand) return -22;
			tmp = operand;
			while (*tmp && (*tmp!=' ') && (*tmp!='\t')) tmp++;	
			*tmp=0;
			if ((tmp==operand) || ((tmp-operand)>MAX_LEN_SYMBOL)) return -24;
			normalize_symbol(operand);
			if (strcmp(module_name, operand)) return -25;	// mismatch error
			module_size = *address - module_start;
			return -27;	// module complete
		}
		else if (!strcmp(opcode, "ORG")) {	// set address counter
			if ((res = get_integer (operand, strlen(operand), &location)) < 0) return res;
			if ((pass == 3) && *module_name) fprintf(stderr, "Line %d: Warning: ORG instruction after NAM instruction.\n", line);
			*address = location;
		}
		else if (!strcmp(opcode, "BSS")) {	// memory block reservation
			if ((res = get_integer (operand, strlen(operand), &size)) < 0) return res;
			if (pass < 3) (*address) += size;
			else 
				for (i=0; i<size; i++) {
					if (i==0) word_out(ofp, *address, 0, label, opcode, operand, comment);
						else word_out(ofp, *address, 0, 0, 0, 0, 0);
					(*address)++;
				}
		}
		else if (!strcmp(opcode, "LIT")) {	// literal & link memory reservation
			if (!operand) return -22;
			if ((res = evaluate_expression(*address, operand, strlen(operand), &element)) < 0) return res;
			if (pass == 1) add_literalspace(*address, element.value);
			else if (pass == 3) literalspace_out(ofp, *address, 1, label, opcode, operand, comment);
			*address += element.value;
		}
		else if (!strcmp(opcode, "DAT")) {	// data generator
			if ((res = data_generator(ofp, pass, address, label, opcode, operand, comment)) < 0) return res;
		}
		else if ((!strcmp(opcode, "EQU")) || (!strcmp(opcode, "SET"))) {	// symbol definitions
			if (repeat_opcode > 1) return -21;
			if (operand) {
				if (pass > 1) {
					if ((res = evaluate_expression(*address, operand, strlen(operand), &element)) < 0) return res;
					if ((i = lookup_symbol(label)) < 0)
						return -10;
					else {
						if (symbol_table[i].vtype == STRING)
							if (symbol_table[i].string) free(symbol_table[i].string);
						symbol_table[i].type = element.type;
						symbol_table[i].vtype = element.vtype;
						if (element.vtype == STRING) {
							symbol_table[i].string = strdup(element.string);
							symbol_table[i].value = 0;
							symbol_table[i].bcd = 0;
						}
						else {
							symbol_table[i].value = element.value;
							symbol_table[i].bcd = element.bcd;
							symbol_table[i].string = 0;
						}
					}
				}
			}
			else return -22;
		}
		else if (!strcmp(opcode, "ENT")) {	// entry point
			if (repeat_opcode > 1) return -21;
			;	// noting to do			
		}
		else if (!strcmp(opcode, "EXT")) {	// external symbol declaration
			return -19;	// not yet supported			
		}
		else if (!strcmp(opcode, "IOF")) {	// turn off automatic indirection
			automatic_indirection = 0;
		}
		else if (!strcmp(opcode, "ION")) {	// turn on automatic indirection
			automatic_indirection = 1;
		}
		else if (!strcmp(opcode, "BON")) {	// turn on base page addressing
			base_page_mode = 1;				// (custom pseudo code)
		}
		else if (!strcmp(opcode, "BOF")) {	// turn off base page addressing
			base_page_mode = 0;				// (custom pseudo code)
		}
		else if (!strcmp(opcode, "REP")) {	// repeate next opcode
			if (repeat_opcode > 1) return -21;
			if (operand) {
				if ((res = evaluate_expression(*address, operand, strlen(operand), &element)) < 0) return res;
			}
			else return -22;
			if (element.vtype == STRING || element.vtype == REAL || element.vtype == SHORT) return -9;
			tmp_repeat_opcode = element.value;
		}

		// remaining system '45 assembler pseudo-instructions (ignored)
		else if ((!strcmp(opcode, "ANY"))||	// any type declaration
				(!strcmp(opcode, "COM")) ||	// basic common declaration
				(!strcmp(opcode, "FIL")) ||	// file number declaration
				(!strcmp(opcode, "HED")) ||	// top-of-page source listing control
				(!strcmp(opcode, "IFA")) ||	// conditional assembly
				(!strcmp(opcode, "IFB")) ||	// conditional assembly
				(!strcmp(opcode, "IFC")) ||	// conditional assembly
				(!strcmp(opcode, "IFD")) ||	// conditional assembly
				(!strcmp(opcode, "IFE")) ||	// conditional assembly
				(!strcmp(opcode, "IFF")) ||	// conditional assembly
				(!strcmp(opcode, "IFH")) ||	// conditional assembly
				(!strcmp(opcode, "IFP")) ||	// conditional assembly
				(!strcmp(opcode, "INT")) ||	// integer declaration
				(!strcmp(opcode, "LST")) ||	// enable source listing
				(!strcmp(opcode, "REL")) ||	// full-precision declaration
				(!strcmp(opcode, "SHO")) ||	// short-precision declaration
				(!strcmp(opcode, "SKP")) ||	// top-of-page source listing control
				(!strcmp(opcode, "SPC")) ||	// blank lines source listing control
				(!strcmp(opcode, "STR")) ||	// string declaration
				(!strcmp(opcode, "SUB")) ||	// subroutine entry point
				(!strcmp(opcode, "UNL")) ||	// disable source listing
				(!strcmp(opcode, "XIF")))	// conditional assembly
			{ if (pass == 3) fprintf(stderr, "Line %d: Warning: unsupported system '45 assembler pseudo instruction.\n", line); }
		
		// remaining ASMA pseudo-instructions (ignored)
		else if ((!strcmp(opcode, "IFN"))|| // conditional assembly
				(!strcmp(opcode, "IFZ")) ||	// conditional assembly
				(!strcmp(opcode, "ORR")) ||	// origin address reset
				(!strcmp(opcode, "DEF")) ||	// address definition
				(!strcmp(opcode, "ABS")) ||	// 16-bit word data
				(!strcmp(opcode, "ASC")) ||	// string data
				(!strcmp(opcode, "DEC")) ||	// signed integer data (decimal)
				(!strcmp(opcode, "OCT")) ||	// unsigned integer data (octal)
				(!strcmp(opcode, "SUP")) ||	// suppress additional listing lines
				(!strcmp(opcode, "UNS"))) 	// re-enable additional listing lines
			{ if (pass == 3) fprintf(stderr, "Line %d: Warning: unsupported ASMA pseudo instruction.\n", line); }

		// real machine opcodes
		else {
			res = code_generator(ofp, pass, *address, label, opcode, operand, comment, line);
			if (res == -18) return res;	// don't increment counter for invalid opcodes
			(*address)++;
			if (res < 0) {
				if (pass == 3) {
					word_out(ofp, *address, 0, 0, 0, 0, comment);	// NOP as placeholder
				}
				return res;
			}
		}
		repeat_opcode--;
		if (strcmp(opcode, "REP")) label = 0;
	} // repeat_opcode
	
	if (listing && (pass == 3) && !opcode) {
		if (numbers) printf("%3d ", listing_line++);
		if (comment) printf(";%s\n", comment);
		else printf("\n");
	}
	if (repeat_opcode == 0) repeat_opcode = 1;
	if (tmp_repeat_opcode > 0) repeat_opcode = tmp_repeat_opcode;
	return 0;
}

int getline(FILE* fp, char* str, int max, char delim)
{
  char buf;
  int i;
  int bytes_read;
  
  i=0;
  while ((i<max)&& ((bytes_read = fread((void *)&buf, 1, 1, fp)) > 0) && (buf != delim)) {
        if (ferror(fp)) {
   	      fprintf(stderr, "Read error on input file.\n");
   	      exit(-1);
   	}
   	if ((buf > 31) || (buf == '\t')) {  // skip non-important ascii control codes
   	    str[i]=buf;
   	    i++;
   	}
  }
  
  str[i]=0;
  if (bytes_read > 0) return i;      /* size of string */
  else return -1;                    /* EOF */
}

int assemble_module(FILE *ifp, FILE *override_fp, int *seek_start, int *start, int *input_line)
{
  int i, j, line, res, all_ok;
  int end_of_file = 0;
  unsigned int address;
  unsigned char str[1024];
  FILE *ofp;
  Literal *literal, *tmp_literal;

  *module_name = 0;				// empty string
  module_start = *start;
  module_size = 0;
  longest_symbol_name = 0;
  ofp = 0;
  
  /* --------------- pass 1: collect symbols & create memory map --------------- */

  address = *start;
  line = *input_line;
  all_ok = 1;

  if (fseek(ifp, *seek_start, SEEK_SET) < 0) {
   	   fprintf(stderr, "Seek error #%d on input file at position %d\n", errno, 0);
   	   exit(-1);
  }

  do {
  	i = getline(ifp, str, MAX_LEN_LINE, '\n');
	if ((res = assemble_line(ofp, 1, &address, str, line)) < 0) {
    	repeat_opcode = 1;
		if (res != -27) all_ok = 0;
      	// print_error(line, res);
	}
   	line++;
  } while ((i >= 0) && (res != -27));	// EOF or end-of-module
  
  // if module not empty (valid code)
  if (address > *start) {
	  
	if (module_size == 0) {
		fprintf(stderr, "Warning: END statement missing.\n");
		module_size = address - module_start;
	  }
	
	/* if not override, open separate output file for module */
	if (override_fp == 0) {
	  	if (*module_name) {					// not empty
	  		fprintf(stderr, "Assembling module \"%s\"", module_name);
	  		strcpy(str, module_name);
	  		strcat(str, ".obj");
		    ofp = fopen(str, "wb");
		}
		else {
			fprintf(stderr, "Warning: NAM statement missing. Using default output file \"out.obj\".\n");
	  		fprintf(stderr, "Assembling module");
	  		strcpy(str, "out.obj");
	  		if (*input_line == 1)
		    	ofp = fopen(str, "wb");	// first module: create
	  		else
		    	ofp = fopen(str, "ab");	// sequal: append
		}
	    if (!ofp) { fprintf(stderr, "Could not open file \"%s\" for writing.\n", str); exit(-1); }
	}
	else {
	  	if (!*module_name) {
	  		fprintf(stderr, "Warning: NAM statement missing.\n");
	  		fprintf(stderr, "Assembling module");
	  	}
	  	else
	  		fprintf(stderr, "Assembling module \"%s\"", module_name);
	  	ofp = override_fp;
	}
	
	if (octal)
	  	fprintf(stderr, " for starting address %06oB ...\n", module_start);
	else 
	  	fprintf(stderr, " for starting address $%04x ...\n", module_start);
	  	
	// add main literal space
	add_literalspace(address, 512);
		  
	if (verbose) {
	  	printf("\n-- literal space table -----------------------------\n");
		  
		for (i=0; i<literalspaces_size; i++) {
		if (octal)
			printf("(%d): address=%06o size=%d\n", i, literalspace_table[i].address, literalspace_table[i].size);
		else
			printf("(%d): address=%04x size=%d\n", i, literalspace_table[i].address, literalspace_table[i].size);
		}
	}
	
	if (all_ok) fprintf(stderr, "Pass 1 completed without errors.\n");
	else fprintf(stderr, "Pass 1 completed with errors!\n");
	
	/* --------------- pass 2: setup literal pools --------------- */
	
	address = *start;
	line = *input_line;
	all_ok = 1;
	*module_name = 0;
	  
	if (fseek(ifp, *seek_start, SEEK_SET) < 0) {
		fprintf(stderr, "Seek error on image file #%d at position %d\n", errno, 0);
		exit(-1);
	}
	
	do {
		i = getline(ifp, str, 1024, '\n');
		if ((res = assemble_line(ofp, 2, &address, str, line)) < 0) {
			repeat_opcode = 1;
			if (res != -27) all_ok = 0;
			// print_error(line, res);
		}
		line++;
	} while ((i >= 0) && (res != -27));
	
	if (symbols) {
		printf("\n-- symbol table ------------------------------------\n");
		for (i=num_predefined; i<symbols_size; i++) {
		    printf("name=\"%s\" ", symbol_table[i].name);
		    if (symbol_table[i].vtype == STRING) {
		    	printf("value=\"%s\" ", symbol_table[i].string);
		    }
		    else {
		    	if (octal)
		    		printf("value=<%oB> ", symbol_table[i].value);
		    	else
		    		printf("value=<$%x> ", symbol_table[i].value);	    		
		    }
		    switch (symbol_table[i].type) {
		    	case ABS: printf("type=<ABS> "); break;
		    	case REL: printf("type=<REL> "); break;
		    	case EXT: printf("type=<EXT> "); break;
		    	default: printf("-unknown type- \n");
		 	}
		    switch (symbol_table[i].vtype) {
		    	case ADDRESS: printf("vtype=<ADDRESS>\n"); break;
		    	case INTEGER: printf("vtype=<INTEGER>\n"); break;
		    	case STRING: printf("vtype=<STRING>\n"); break;
		     	case ASCII: printf("vtype=<ASCII>\n"); break;
		    	case REAL: printf("vtype=<REAL>\n"); break;
		    	case SHORT: printf("vtype=<SHORT>\n"); break;
		   	default: printf("-unknown vtype-\n");
		 	}
		}
		printf("\n");
	}
	  
	if (verbose) {
		printf("\n-- literal pool ------------------------------------\n");
		for (i=0; i<literalspaces_size; i++) {
		if (octal)
			printf("(%d): address=%06o size=%d used=%d\n", i, literalspace_table[i].address, literalspace_table[i].size, literalspace_table[i].used);
		else
			printf("(%d): address=%04x size=%d used=%d\n", i, literalspace_table[i].address, literalspace_table[i].size, literalspace_table[i].used);    	
		literal = literalspace_table[i].list;
		while (literal) {
			if (octal)
				printf("     address=%06o size=%d", literal->address, literal->size);
			else
				printf("     address=%04x size=%d", literal->address, literal->size);    		
			for (j=0; j<literal->size; j++) {
				if (octal)
					printf(" %06o", literal->value[j*2]*256 + literal->value[j*2+1]);
				else
					printf(" %04x", literal->value[j*2]*256 + literal->value[j*2+1]);
			}
			printf("\n");
			literal = literal->next;
		}
		}
	}
	  
	if (all_ok) fprintf(stderr, "Pass 2 completed without errors.\n");
	else fprintf(stderr, "Pass 2 completed with errors!\n");
	
	/* --------------- pass 3: generating machine code --------------- */
	  
	address = *start;
	line = *input_line;						// source line number
	all_ok = 1;
	*module_name = 0;
	  
	if (fseek(ifp, *seek_start, SEEK_SET) < 0) {
	   	   fprintf(stderr, "Seek error on image file #%d at position %d\n", errno, 0);
	   	   exit(-1);
	}
	  
	do {
		i = getline(ifp, str, 1024, '\n');
		if ((res = assemble_line(ofp, 3, &address, str, line)) < 0) {
			repeat_opcode = 1;
			if (res != -27) {
				print_error(line, res);
				all_ok = 0;
			}
		}
		line++;
	} while ((i >= 0) && (res != -27));
	  
	if (all_ok) fprintf(stderr, "Pass 3 completed without errors.\n");
	else fprintf(stderr, "Pass 3 completed with errors!\n");
  	
  }
  	
  if (i < 0) end_of_file = 1;
  
  /* append default literal space at the end of the module (special case) */
  literalspace_out(ofp, address, 0, 0, 0, 0, 0);
  address += literalspace_table[literalspaces_size-1].used;
  
  /* update counters */
  *seek_start = ftell(ifp);
  *start = address;
  *input_line = line;
  
  /* cleanup symbol table */
  for (i=num_predefined; i<symbols_size; i++) {
    if (symbol_table[i].name) free(symbol_table[i].name);
    if (symbol_table[i].vtype == STRING) {
    	if (symbol_table[i].string) free(symbol_table[i].string);
    }
	symbol_table[i].type = 0;
	symbol_table[i].vtype = 0;
	symbol_table[i].value = 0;
	symbol_table[i].string = 0;
  }
  symbols_size = num_predefined;

  /* cleanup literals */
  for (i=0; i<literalspaces_size; i++) {
    literal = literalspace_table[i].list;
    while (literal) {
    	tmp_literal = literal;
    	literal = literal->next;
		free(tmp_literal->value);
    	free(tmp_literal);
    }
  }
  literalspaces_size = 0;
  
  if (!override_fp && ofp) fclose(ofp);
  	
  if (end_of_file) return 0; 	// end-of-file reached, nothing more to do
  else return 1; 				// not yet finished
}
